//
// fraktal - visualization of Mandelbrot and Julia sets
// Copyright (C) 1999-2020  Joachim Reichel <joachim.reichel@posteo.de>
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//

#include <iostream>
#include <cmath>

#if defined( HAVE_GLEW)
#  include <GL/glew.h>
#endif // HAVE_GLEW

#if defined( HAVE_QT)
#  include <QtWidgets/QApplication>
#  include <QtWidgets/QMessageBox>
#  include <QtOpenGL/QGLFormat>
#endif // HAVE_QT

#include "mainwindow.h"
#include "renderer.h"

int main( int argc, char* argv[])
{
    Function func = FCT_MANDELBROT;
    Mode mode = MODE_CPU;
    bool fullscreen = false;
    bool screenshot = false;
    bool center = false;
    bool scaling = false;
    bool quality = false;
    double cx = 0.0;
    double cy = 0.0;
    double s  = 1.0;
    int q = 0;
    int width = 1600;
    int height = 1200;
    std::string filename;

    for( int i=1; i<argc; i++) {

        if( (strcmp( argv[i], "--help") == 0) ||( strcmp( argv[i], "-h") == 0)) {
            std::cout << "Usage: " << FRAKTAL_PACKAGE << " [-h|--help] [-v|--version]" << std::endl;
            std::cout << "       " << FRAKTAL_PACKAGE << " [-m|--mandelbrot] [-i|--inverse] [-j|--julia] \\" << std::endl;
            std::cout << "               [-c|--center x y] [-s|--scaling s] [-q|--quality q] \\" << std::endl;
            std::cout << "               [-o|--opengl] [-d|--cuda] [-f|--fullscreen] \\" << std::endl;
            std::cout << "               [--screenshot w h filename]" << std::endl;
            std::cout << std::endl;
            std::cout << "  -h --help        display this help message" << std::endl;
            std::cout << "  -v --version     display version number" << std::endl;
            std::cout << std::endl;
            std::cout << "  -m --mandelbrot  select mandelbrot function (default)" << std::endl;
            std::cout << "  -i --inverse     select inverse mandelbrot function" << std::endl;
            std::cout << "  -j --julia       select julia function" << std::endl;
            std::cout << std::endl;
            std::cout << "  -c --center x y  center image at x+yi" << std::endl;
            std::cout << "  -s --scaling s   scale image by factor s > 0" << std::endl;
            std::cout << "  -q --quality q   uses q x q samples per pixel" << std::endl;
            std::cout << std::endl;
            std::cout << "  -o --opengl      use OpenGL mode (if available)" << std::endl;
            std::cout << "  -d --cuda        use CUDA mode (if available)" << std::endl;
            std::cout << "  -f --fullscreen  start in fullscreen mode" << std::endl;
            std::cout << std::endl;
            std::cout << "  --screenshot w h filename    save image of size w x h as PNG" << std::endl;
            std::cout << std::endl;
            return 1;

        } else if( (strcmp( argv[i], "--version") == 0) || (strcmp( argv[i], "-v") == 0)) {
            std::cout << FRAKTAL_PACKAGE << " " << FRAKTAL_VERSION << std::endl;
            return 2;

        } else if( (strcmp( argv[i], "--mandelbrot") == 0) || (strcmp( argv[i], "-m") == 0)) {
            func = FCT_MANDELBROT;

        } else if( (strcmp( argv[i], "--inverse") == 0) || (strcmp( argv[i], "-i") == 0)) {
            func = FCT_INVERSE_MANDELBROT;

        } else if( (strcmp( argv[i], "--julia") == 0) || (strcmp( argv[i], "-j") == 0)) {
            func = FCT_JULIA;

        } else if( ((strcmp( argv[i], "--center") == 0) || (strcmp( argv[i], "-c") == 0)) && (i<argc-2)) {
            center = true;
            cx = atof( argv[++i]);
            cy = atof( argv[++i]);

        } else if( ((strcmp( argv[i], "--scaling") == 0) || (strcmp( argv[i], "-s") == 0)) && (i<argc-1)) {
            scaling = true;
            s  = fabs( atof( argv[++i]));
            if( s == 0)
                s = 1;

        } else if( ((strcmp( argv[i], "--quality") == 0) || (strcmp( argv[i], "-q") == 0)) && (i<argc-1)) {
            quality = true;
            q = atoi( argv[++i]);

        } else if( (strcmp( argv[i], "--opengl") == 0) || (strcmp( argv[i], "-o") == 0)) {
            mode = MODE_OPENGL;

        } else if( (strcmp( argv[i], "--cuda") == 0) ||( strcmp( argv[i], "-d") == 0)) {
            mode = MODE_CUDA;

        } else if( (strcmp( argv[i], "--fullscreen") == 0) || (strcmp( argv[i], "-f") == 0)) {
            fullscreen = true;

        } else if( (strcmp( argv[i], "--screenshot") == 0) && (i<argc-3)) {
            screenshot = true;
            int w = atoi( argv[++i]);
            int h = atoi( argv[++i]);
            if( w > 0)
                width = w;
            if( h > 0)
                height = h;
            filename = std::string( argv[++i]);

        } else {
            std::cout << "Unknown option: \"" << argv[i] << "\"" << std::endl;
        }
    }

#if defined( HAVE_QT)

    try {

        // Create Qt application
        QApplication* application = new QApplication( argc, argv);

        if( !QGLFormat::hasOpenGL())  {
            QMessageBox::critical( NULL, "Error", "This system has no OpenGL support. Exiting.");
            return 3;
        }

        // Create top level widget
        MainWindow* mainwindow = new MainWindow();
        Renderer* renderer = mainwindow->get_renderer();
        renderer->set_function( func);
        renderer->set_mode( mode);
        renderer->reset();

        // Handle options
        if( fullscreen)
            mainwindow->misc_fullscreen();
        if( center)
            renderer->set_center( cx, cy);
        if( scaling)
            renderer->set_scaling( s);
        if( quality)
            renderer->set_quality( q);

        if( screenshot) {
#if defined( HAVE_PNG) || defined( HAVE_JPEG)
            renderer->screenshot( width, height, filename);
            delete mainwindow;
            delete application;
            return 0;
#else // HAVE_PNG || HAVE_JPEG
            (void) width;  // avoid warning
            (void) height; // avoid warning
            std::cerr << "Screenshots need PNG or JPEG support. Exiting.";
            delete mainwindow;
            delete application;
            return 4;
#endif // HAVE_PNG || HAVE_JPEG
        }

        mainwindow->show();
        int exitcode = application->exec();

        delete mainwindow;
        delete application;

        return exitcode;

    } catch( const std::exception& e) {

       std::cerr << "Exception: " << e.what() << std::endl;
       return 5;

    }

#else // HAVE_QT

    std::cerr << "This program needs Qt." << std::endl;
    return 6;

#endif // HAVE_QT
}
