//
// fraktal - visualization of Mandelbrot and Julia sets
// Copyright (C) 1999-2020  Joachim Reichel <joachim.reichel@posteo.de>
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//

#include "glwidget.h"

#include <cmath>

#include <QtGui/QKeyEvent>
#include <QtGui/QMouseEvent>

#include "mainwindow.h"
#include "renderer.h"

const int lowest_quality  = -5;
const int highest_quality =  2;

GLWidget::GLWidget( MainWindow* mainwindow, QWidget* parent)
  : QOpenGLWidget( parent), m_mainwindow( mainwindow), m_quality( 0)
{
    setFocusPolicy( Qt::ClickFocus);
}

void GLWidget::initializeGL()
{
    m_mainwindow->get_renderer()->init();
}

void GLWidget::paintGL()
{
    m_mainwindow->get_renderer()->draw( width(), height());
}

void GLWidget::resizeGL( int w, int h)
{
    m_mainwindow->get_renderer()->resize( w, h);
}

void GLWidget::mousePressEvent( QMouseEvent* e)
{
    m_button_state = e->button();
    m_last = e->pos();

    double w = width();
    double h = height();
    double d = std::min( w, h) / 2.0;

    m_px =  ( m_last.x() - w/2.0) / d;
    m_py = -( m_last.y() - h/2.0) / d;

    m_mainwindow->get_renderer()->set_quality( m_quality);
}

void GLWidget::mouseMoveEvent( QMouseEvent* e)
{
    Renderer* renderer = m_mainwindow->get_renderer();
    QPoint current = e->pos();

    double w = width();
    double h = height();
    double d = std::min( w, h) / 2.0;

    if( m_button_state != Qt::MidButton) {
        double dx = -( current.x() - m_last.x()) / d;
        double dy =  ( current.y() - m_last.y()) / d;
        renderer->translate( dx, dy);
    } else {
        double dy = ( current.y() - m_last.y()) / d;
        double s = exp( 2.0 * dy);
        renderer->scale( s, m_px, m_py);
    }


    std::chrono::system_clock::time_point start = std::chrono::system_clock::now();

    update();

    std::chrono::system_clock::time_point stop  = std::chrono::system_clock::now();
    double duration = std::chrono::duration<double>( stop-start).count();

    if( (duration > 0.12) && (m_quality > lowest_quality)) {
        m_quality--;
        renderer->set_quality( m_quality);

    } else if( (duration < 0.04) && (m_quality < highest_quality)) {
        m_quality++;
        renderer->set_quality( m_quality);
    }

    m_mainwindow->message_tmp( renderer->status( width(), height()).c_str());

    m_last = current;
}

void GLWidget::mouseReleaseEvent( QMouseEvent* /*e*/)
{
    Renderer* renderer = m_mainwindow->get_renderer();
    m_button_state = Qt::NoButton;
    renderer->set_quality( highest_quality);
    m_mainwindow->message_tmp( renderer->status( width(), height()).c_str());
    update();
}

void GLWidget::keyReleaseEvent( QKeyEvent* e)
{
    e->ignore();
}
