//
// fraktal - visualization of Mandelbrot and Julia sets
// Copyright (C) 1999-2020  Joachim Reichel <joachim.reichel@posteo.de>
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//

#include "mainwindow.h"

#include <cassert>

#if defined( HAVE_GLEW)
#  include <GL/glew.h>
#endif // HAVE_GLEW

#include <QtGui/QKeyEvent>
#include <QtWidgets/QApplication>
#include <QtWidgets/QFileDialog>
#include <QtWidgets/QFrame>
#include <QtWidgets/QHBoxLayout>
#include <QtWidgets/QMenuBar>
#include <QtWidgets/QMessageBox>
#include <QtWidgets/QStatusBar>
#include <QtWidgets/QOpenGLWidget>

#include "glwidget.h"
#include "renderer.h"

MainWindow::MainWindow()
  : QMainWindow()
{
    resize( 800, 600);

    // create Renderer
    m_renderer = new Renderer();

    // frame
    QFrame* frame = new QFrame( this);
    setCentralWidget( frame);

    // OpenGL widget
    m_glwidget = new GLWidget( this, frame);

    // horizontal layout
    QHBoxLayout* flayout = new QHBoxLayout( frame);
    flayout->addWidget( m_glwidget);

    setFocusPolicy( Qt::StrongFocus);

    m_menu_function  = new QMenu( "Function", this);
    m_menu_mode      = new QMenu( "Mode", this);
    m_menu_misc      = new QMenu( "Misc", this);

    connect( m_menu_function, SIGNAL( aboutToShow()), this, SLOT( update_menus()));
    connect( m_menu_mode,     SIGNAL( aboutToShow()), this, SLOT( update_menus()));
    connect( m_menu_misc,     SIGNAL( aboutToShow()), this, SLOT( update_menus()));

    menuBar()->addMenu( m_menu_function);
    menuBar()->addMenu( m_menu_mode);
    menuBar()->addMenu( m_menu_misc);

    m_action_function_mandelbrot_set = m_menu_function->addAction(
        "&Mandelbrot set", this, SLOT( function_mandelbrot_set()), QKeySequence( Qt::Key_M));
    m_action_function_inverse_mandelbrot_set = m_menu_function->addAction(
        "&Inverse Mandelbrot set", this, SLOT( function_inverse_mandelbrot_set()), QKeySequence( Qt::Key_I));
    m_action_function_julia_set = m_menu_function->addAction(
        "&Julia set", this, SLOT( function_julia_set()), QKeySequence( Qt::Key_J));

    m_action_mode_cpu = m_menu_mode->addAction(
        "&CPU", this, SLOT( mode_cpu()), QKeySequence( Qt::Key_C));
    m_action_mode_opengl = m_menu_mode->addAction(
        "&OpenGL", this, SLOT( mode_opengl()), QKeySequence( Qt::Key_O));
    m_action_mode_cuda = m_menu_mode->addAction(
        "CU&DA", this, SLOT( mode_cuda()), QKeySequence( Qt::Key_D));

    m_action_misc_reset = m_menu_misc->addAction(
        "&Reset", this, SLOT( misc_reset()), QKeySequence( Qt::Key_R));
    m_action_misc_screenshot = m_menu_misc->addAction(
        "&Screenshot", this, SLOT( misc_screenshot()), QKeySequence( Qt::Key_S));
    m_action_misc_fullscreen = m_menu_misc->addAction(
        "&Fullscreen", this, SLOT( misc_fullscreen()), QKeySequence( Qt::Key_F));
    m_action_misc_about = m_menu_misc->addAction(
        "&About", this, SLOT( misc_about()), QKeySequence( Qt::Key_A));
    m_menu_misc->addSeparator();
    m_action_misc_quit = m_menu_misc->addAction(
        "&Quit", QApplication::instance(), SLOT( quit()), QKeySequence( Qt::Key_Q));

    m_action_function_mandelbrot_set        ->setCheckable( true);
    m_action_function_inverse_mandelbrot_set->setCheckable( true);
    m_action_function_julia_set             ->setCheckable( true);

    m_action_mode_cpu                       ->setCheckable( true);
    m_action_mode_opengl                    ->setCheckable( true);
    m_action_mode_cuda                      ->setCheckable( true);

    m_action_misc_fullscreen                ->setCheckable( true);

#if !defined( HAVE_PNG) && !defined( HAVE_JPEG)
    m_action_misc_screenshot->setEnabled( false);
#endif // HAVE_PNG || HAVE_JPEG

#if defined( HAVE_GLEW)
    m_action_mode_opengl->setEnabled( true);
#else
    m_action_mode_opengl->setEnabled( false);
#endif // HAVE_GLEW

#if defined( HAVE_CUDA)
    m_action_mode_cuda->setEnabled( true);
#else
    m_action_mode_cuda->setEnabled( false);
#endif // HAVE_CUDA

    message_tmp( "Ready");
}

void MainWindow::update_menus()
{
    m_action_function_mandelbrot_set        ->setChecked( m_renderer->get_function() == FCT_MANDELBROT);
    m_action_function_inverse_mandelbrot_set->setChecked( m_renderer->get_function() == FCT_INVERSE_MANDELBROT);
    m_action_function_julia_set             ->setChecked( m_renderer->get_function() == FCT_JULIA);

    m_action_mode_cpu                       ->setChecked( m_renderer->get_mode() == MODE_CPU);
    m_action_mode_opengl                    ->setChecked( m_renderer->get_mode() == MODE_OPENGL);
    m_action_mode_cuda                      ->setChecked( m_renderer->get_mode() == MODE_CUDA);

    m_action_misc_fullscreen                ->setChecked( windowState() & Qt::WindowFullScreen);
}

void MainWindow::function_mandelbrot_set()
{
    m_renderer->set_function( FCT_MANDELBROT);
    m_glwidget->update();
}

void MainWindow::function_inverse_mandelbrot_set()
{
    m_renderer->set_function( FCT_INVERSE_MANDELBROT);
    m_glwidget->update();
}

void MainWindow::function_julia_set()
{
    m_renderer->set_function( FCT_JULIA);
    m_glwidget->update();
}

void MainWindow::mode_cpu()
{
    m_renderer->set_mode( MODE_CPU);
    m_glwidget->update();
}

void MainWindow::mode_opengl()
{
    m_renderer->set_mode( MODE_OPENGL);
    m_glwidget->update();
}

void MainWindow::mode_cuda()
{
    m_renderer->set_mode( MODE_CUDA);
    m_glwidget->update();
}

void MainWindow::misc_reset()
{
    m_renderer->reset();
    message_tmp( m_renderer->status( m_glwidget->width(), m_glwidget->height()).c_str());
    m_glwidget->update();
}

void MainWindow::misc_screenshot()
{
#if defined( HAVE_PNG) || defined( HAVE_JPEG)
    QString filter = "Images( ";
#if defined( HAVE_PNG)
    filter += "*.png ";
#else // HAVE_PNG
    filter += "*.jpg *.jpeg ";
#endif // HAVE_PNG
    filter.remove( filter.length()-1, 1);
    filter += ")";

    QString selected_filter;
    QString functionname = QFileDialog::getSaveFileName( this, "", ".", filter, &selected_filter);
    if( functionname == "")
        return;

    message( "Computing screenshot ...");
    std::string s( functionname.toLatin1());
    m_renderer->screenshot( 1600, 1200, s);
    message_tmp( "Computing screenshot done.");
#else // HAVE_PNG || HAVE_JPEG
    assert( false);
#endif // HAVE_PNG || HAVE_JPEG
}

void MainWindow::misc_fullscreen()
{
    setWindowState( windowState() ^ Qt::WindowFullScreen);
}

void MainWindow::misc_about()
{
    QMessageBox::information( this, "About", QString( "%1 %2\n\n(c) 1999-2020  Joachim Reichel\n    <joachim.reichel@posteo.de>").
        arg(FRAKTAL_PACKAGE).arg(FRAKTAL_VERSION), QMessageBox::Ok);
}

Renderer* MainWindow::get_renderer() const
{
    return m_renderer;
}

const GLWidget* MainWindow::get_glwidget() const
{
    return m_glwidget;
}

void MainWindow::keyPressEvent( QKeyEvent* e)
{
    if( e->key() == Qt::Key_Escape)
        QApplication::instance()->quit();
    else
        QMainWindow::keyPressEvent( e);
}

void MainWindow::keyReleaseEvent( QKeyEvent* e)
{
    QMainWindow::keyPressEvent( e);
}

void MainWindow::message( const QString & message) const
{
    statusBar()->showMessage( message);
}

void MainWindow::message_tmp( const QString & message) const
{
    statusBar()->showMessage( message, 5000);
}

void MainWindow::message_clear() const
{
    statusBar()->clearMessage();
}
