//
// fraktal - visualization of Mandelbrot and Julia sets
// Copyright (C) 1999-2020  Joachim Reichel <joachim.reichel@posteo.de>
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//

#include "renderer.h"

#include <cassert>
#include <cmath>
#include <iostream>
#include <memory>
#include <sstream>
#include <stdexcept>

#if defined( HAVE_GLEW)
#  include <GL/glew.h>
#endif // HAVE_GLEW

#include <GL/glu.h>

#if defined( HAVE_JPEG)
#  include <cstdio>
extern "C" {
#  include <jpeglib.h>
}
#endif // HAVE_JPEG

#if defined( HAVE_PNG)
extern "C" {
#  include <png.h>
}
#endif // HAVE_PNG

#include "shader_vertex.h"
#include "shader_mandelbrot.h"
#include "shader_inverse_mandelbrot.h"
#include "shader_julia.h"

#include "compute.h"

Renderer::Renderer()
  : m_shaders_enabled( false)
{
    set_function( FCT_MANDELBROT);
    set_mode( MODE_CPU);
    reset();
}

Renderer::~Renderer() // NOLINT modernize-use-equals-default
{
#if defined( HAVE_GLEW)
    disable_shaders();
#endif // HAVE_GLEW
}

void Renderer::init()
{
    glDisable( GL_SCISSOR_TEST);
    glDisable( GL_ALPHA_TEST);
    glDisable( GL_STENCIL_TEST);
    glDisable( GL_DEPTH_TEST);
    glDisable( GL_BLEND);
    glDisable( GL_DITHER);
    glDisable( GL_COLOR_LOGIC_OP);
    glDisable( GL_INDEX_LOGIC_OP);
    glDisable( GL_COLOR_MATERIAL);

#if defined( HAVE_GLEW)
    glewInit();

    if( m_mode == MODE_OPENGL)
        enable_shaders();
#endif // HAVE_GLEW
}

void Renderer::draw( int width, int height)
{
    glClear( GL_COLOR_BUFFER_BIT);

#if defined( HAVE_GLEW)
    if( m_mode == MODE_OPENGL && !m_shaders_enabled)
        enable_shaders();

    if( m_mode != MODE_OPENGL && m_shaders_enabled)
        disable_shaders();

    m_kmax = 400 *( 1 + log( m_s) / log(10));

    if( m_shaders_enabled) {

        if( m_fragment_shader != m_shader)
            switch_shaders();

        glLoadIdentity();
        glScaled( m_s, m_s, m_s);
        glTranslated( -m_cx, -m_cy, 0.0);
        glClear( GL_DEPTH_BUFFER_BIT);
        glColor3d( 1.0, 1.0, 1.0);

        double w = width;
        double h = height;
        // compute extend in world coordinates
        double dx =( w < h) ? 1/m_s : 1/m_s * w/h;
        double dy =( h < w) ? 1/m_s : 1/m_s * h/w;
        // compute lower left corner in world coordinates
        double px = m_cx - dx;
        double py = m_cy - dy;
        // compute upper right corner in world coordinates
        double qx = m_cx + dx;
        double qy = m_cy + dy;

        glBegin( GL_QUADS);
        glVertex3d( px, py, 0.0);
        glVertex3d( qx, py, 0.0);
        glVertex3d( qx, qy, 0.0);
        glVertex3d( px, qy, 0.0);
        glEnd();

    } else {
#endif // HAVE_GLEW

        int stride = 8 *( width/8 + 1);
        auto buffer = std::make_unique<GLfloat[]>( 4*stride*height);
        if( m_mode == MODE_CPU)
            compute_cpu( buffer.get(), width, height, m_cx, m_cy, m_s, m_function, m_quality, stride);
        else if( m_mode == MODE_CUDA)
            compute_cuda( buffer.get(), width, height, m_cx, m_cy, m_s, m_function, m_quality, stride);
        glPixelStorei( GL_UNPACK_ALIGNMENT, 8);
        glDrawPixels( stride, height, GL_RGBA, GL_FLOAT, static_cast<GLvoid*>( buffer.get()));

#ifdef HAVE_GLEW
    }
#endif // HAVE_GLEW
}

#if defined( HAVE_PNG) || defined( HAVE_JPEG)
void Renderer::screenshot( int width, int height, const std::string& filename)
{
    int stride1 = 8 * (width/8 + 1);
    GLfloat* buffer1 = new GLfloat[4*stride1*height];

    if( m_mode == MODE_CUDA)
        compute_cuda( buffer1, width, height, m_cx, m_cy, m_s, m_function, m_quality, stride1);
    else
        compute_cpu( buffer1, width, height, m_cx, m_cy, m_s, m_function, m_quality, stride1);

    int stride2 = width;
    char* buffer2 = new char[3*stride2*height];
    for( int y = 0; y < height; y++)
        for( int x = 0; x < width; x++) {
            int offset1 = 4 *( y*stride1 + x);
            int offset2 = 3 *( y*stride2 + x);
            buffer2[offset2+0] = static_cast<char>( buffer1[offset1+0] * 255);
            buffer2[offset2+1] = static_cast<char>( buffer1[offset1+1] * 255);
            buffer2[offset2+2] = static_cast<char>( buffer1[offset1+2] * 255);
        }
    delete[] buffer1;

#if defined( HAVE_PNG)
    write_png_file( buffer2, width, height, filename);
#elif defined( HAVE_JPEG)
    write_jpeg_file( buffer2, width, height, filename);
#else // HAVE_JPEG
    assert( false);
#endif // HAVE_JPEG

    delete[] buffer2;
}
#endif // HAVE_PNG || HAVE_JPEG

void Renderer::resize( int w, int h) const
{
    glViewport( 0, 0, static_cast<GLsizei>( w), static_cast<GLsizei>( h));
    glMatrixMode( GL_PROJECTION);
    glLoadIdentity();
    double ratio = static_cast<double>( w) / static_cast<double>( h);
    if( ratio>1)
        glOrtho( -ratio, ratio, -1, 1, -10, 10);
    else
        glOrtho( -1, 1, -1/ratio, 1/ratio, -10, 10);
    glMatrixMode( GL_MODELVIEW);
}

void Renderer::reset()
{
    set_center( 0.0, 0.0);
    set_scaling( 0.75);
    set_quality( 0);

    if( m_function == FCT_MANDELBROT)
        set_center( -0.6, 0.0);
}

void Renderer::translate( double dx, double dy)
{
    m_cx += dx / m_s;
    m_cy += dy / m_s;
}

void Renderer::scale( double s)
{
    assert( s > 0);
    m_s *= s;
}

void Renderer::scale( double s, double px, double py)
{
    assert( s > 0);

    double qx = m_cx + px / m_s;
    double qy = m_cy + py / m_s;

    m_s *= s;
    m_cx = qx +( m_cx - qx) / s;
    m_cy = qy +( m_cy - qy) / s;
}

void Renderer::set_quality( int q)
{
    m_quality = q;
}

int Renderer::get_quality() const
{
    return m_quality;
}

void Renderer::set_center( double cx, double cy)
{
    m_cx = cx;
    m_cy = cy;
}

void Renderer::get_center( double& cx, double& cy) const
{
    cx = m_cx;
    cy = m_cy;
}

void Renderer::set_scaling( double s)
{
    assert( s > 0);
    m_s = s;
}

double Renderer::get_scaling() const
{
   return m_s;
}

void Renderer::set_function( Function f)
{
    Function old_function = m_function;

#ifdef HAVE_GLEW
    switch( f) {
        case FCT_MANDELBROT:
            m_shader = m_mandelbrot_shader;
            break;
        case FCT_INVERSE_MANDELBROT:
            m_shader = m_inverse_mandelbrot_shader;
            break;
        case FCT_JULIA:
            m_shader = m_julia_shader;
            break;
        default:
            assert( false);
    }
#endif // HAVE_GLEW

    m_function = f;

    if( m_function != old_function)
        reset();
}

Function Renderer::get_function() const
{
    return m_function;
}

void Renderer::set_mode( Mode m)
{
#ifndef HAVE_GLEW
    if( m == MODE_OPENGL) {
        std::cerr << "OpenGL mode not supported." << std::endl;
        return;
    }
#endif // HAVE_GLEW

#ifndef HAVE_CUDA
    if( m == MODE_CUDA) {
        std::cerr << "CUDA mode not supported." << std::endl;
        return;
    }
#endif // HAVE_CUDA

    m_mode = m;
}

Mode Renderer::get_mode() const
{
    return m_mode;
}

std::string Renderer::status( int width, int height) const
{
    int samples = m_quality >= 0 ? 1 << m_quality : 1;
    int size = m_quality <= 0 ? 1 << -m_quality : 1;

    std::ostringstream str;
    str << "center: (" << m_cx << ", " << m_cy << "), ";
    str << "scaling factor: " << m_s << ", ";
    str << "width: " << width << ", ";
    str << "height: " << height << ", ";
    str << "quality: " << m_quality << " (";
    str << "samples: " << samples <<  "x" << samples << ", ";
    str << "block size: " << size << "x" << size << ")";
    return str.str();
}

#if defined( HAVE_JPEG)
void Renderer::write_jpeg_file( const char* buffer, int width, int height, const std::string& filename) const
{
    struct jpeg_compress_struct cinfo;
    struct jpeg_error_mgr jerr;

    cinfo.err = jpeg_std_error( &jerr);
    jpeg_create_compress( static_cast<j_compress_ptr>( &cinfo));

    FILE* outfile = fopen( filename.c_str(), "wb");
    if( !outfile) {
        std::ostringstream str;
        str << "can't open " << filename;
        throw std::runtime_error( str.str());
    }

    jpeg_stdio_dest( &cinfo, outfile);

    cinfo.image_width = width;
    cinfo.image_height = height;
    cinfo.input_components = 3;
    cinfo.in_color_space = JCS_RGB;

    jpeg_set_defaults( static_cast<j_compress_ptr>( &cinfo));
    jpeg_set_quality( &cinfo, 100, TRUE);
    jpeg_start_compress( &cinfo, TRUE);

    while( cinfo.next_scanline < cinfo.image_height) {
        const char* p = &buffer[ 3 *( cinfo.image_height-cinfo.next_scanline-1) * width ];
        JSAMPROW row_pointer = reinterpret_cast<JSAMPROW>( const_cast<char*>( p));
        jpeg_write_scanlines( &cinfo, &row_pointer, 1);
    }

    jpeg_finish_compress( &cinfo);
    fclose( outfile);
    jpeg_destroy_compress( &cinfo);
}
#endif // HAVE_JPEG

#if defined( HAVE_PNG)
void Renderer::write_png_file( const char* buffer, int width, int height, const std::string& filename) const
{
    FILE* outfile = fopen( filename.c_str(), "wb");
    if( !outfile) {
        std::ostringstream str;
        str << "can't open " << filename;
        throw std::runtime_error( str.str());
    }

    png_structp png_ptr = png_create_write_struct( PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
    if( !png_ptr) {
        fclose( outfile);
        throw std::runtime_error( "can't create png_ptr");
    }

    png_infop info_ptr = png_create_info_struct( png_ptr);
    if( !info_ptr) {
        png_destroy_write_struct( static_cast<png_structpp>( &png_ptr), /*png_infopp_NULL*/ NULL);
        fclose( outfile);
        throw std::runtime_error( "can't create info_ptr");
    }

    if( setjmp( png_jmpbuf( png_ptr))) {
        png_destroy_write_struct( &png_ptr, &info_ptr);
        fclose( outfile);
        throw std::runtime_error( "png runtime error");
    }

    png_init_io( png_ptr, outfile);
    png_set_IHDR( png_ptr, info_ptr, width, height, 8, PNG_COLOR_TYPE_RGB,
        PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_BASE, PNG_FILTER_TYPE_BASE);
    png_write_info( png_ptr, info_ptr);

    png_bytep row_pointers[height];
    for( int k=0; k<height; k++) {
        const char* p = &buffer[ 3 *( height-k-1) * width ];
        row_pointers[k] = reinterpret_cast<png_bytep>( const_cast<char*>( p));
    }

    png_write_image( png_ptr, row_pointers);
    png_write_end( png_ptr, info_ptr);
    png_destroy_write_struct( &png_ptr, &info_ptr);

    fclose( outfile);
}
#endif // HAVE_PNG

#if defined( HAVE_GLEW)
void Renderer::enable_shaders()
{
    if( m_shaders_enabled)
        return;

    if( !glewIsSupported( "GL_VERSION_2_0")) {
        std::cerr << "OpenGL 2.0 not supported, vertex and fragment shaders will be disabled" << std::endl;
        return;
    }

    m_vertex_shader = glCreateShader( GL_VERTEX_SHADER);
    m_mandelbrot_shader = glCreateShader( GL_FRAGMENT_SHADER);
    m_inverse_mandelbrot_shader = glCreateShader( GL_FRAGMENT_SHADER);
    m_julia_shader = glCreateShader( GL_FRAGMENT_SHADER);
    set_function( get_function());

    glShaderSource( m_vertex_shader, 1, &g_vertex_shader_string, NULL);
    glShaderSource( m_mandelbrot_shader, 1, &g_mandelbrot_shader_string, NULL);
    glShaderSource( m_inverse_mandelbrot_shader, 1, &g_inverse_mandelbrot_shader_string, NULL);
    glShaderSource( m_julia_shader, 1, &g_julia_shader_string, NULL);

    glCompileShader( m_vertex_shader);
    print_shader_info_log( m_vertex_shader);

    glCompileShader( m_mandelbrot_shader);
    print_shader_info_log( m_mandelbrot_shader);

    glCompileShader( m_inverse_mandelbrot_shader);
    print_shader_info_log( m_inverse_mandelbrot_shader);

    glCompileShader( m_julia_shader);
    print_shader_info_log( m_julia_shader);

    m_shaderprogram = glCreateProgram();
    print_program_info_log( m_shaderprogram);

    glAttachShader( m_shaderprogram, m_vertex_shader);
    print_program_info_log( m_shaderprogram);

    m_fragment_shader = m_shader;

    glAttachShader( m_shaderprogram, m_fragment_shader);
    print_program_info_log( m_shaderprogram);

    glLinkProgram( m_shaderprogram);
    print_program_info_log( m_shaderprogram);

    glUseProgram( m_shaderprogram);
    print_program_info_log( m_shaderprogram);

    GLuint loc = glGetUniformLocation( m_shaderprogram, "kmax");
    glUniform1fv( loc, 1, &m_kmax);

    m_shaders_enabled = true;
}

void Renderer::switch_shaders()
{
    if( !m_shaders_enabled)
        return;

    glDetachShader( m_shaderprogram, m_fragment_shader);

    m_fragment_shader = m_shader;

    glAttachShader( m_shaderprogram, m_fragment_shader);
    glLinkProgram( m_shaderprogram);
    glUseProgram( m_shaderprogram);

    GLuint loc = glGetUniformLocation( m_shaderprogram, "kmax");
    glUniform1fv( loc, 1, &m_kmax);
}

void Renderer::disable_shaders()
{
    if( !m_shaders_enabled)
        return;

    glUseProgram( 0);

    glDetachShader( m_shaderprogram, m_vertex_shader);
    glDetachShader( m_shaderprogram, m_fragment_shader);

    glDeleteShader( m_vertex_shader);
    glDeleteShader( m_mandelbrot_shader);
    glDeleteShader( m_inverse_mandelbrot_shader);
    glDeleteShader( m_julia_shader);

    glDeleteProgram( m_shaderprogram);

    m_shaders_enabled = false;
}

void Renderer::print_shader_info_log( GLuint obj) const
{
    int length;
    glGetProgramiv( obj, GL_INFO_LOG_LENGTH, &length);
    if( length > 1) {
        char* buffer = new char[length];
        int actual_length;
        glGetShaderInfoLog( obj, length, &actual_length, buffer);
        if( actual_length > 1)
            std::cerr << buffer << std::endl;
        delete[] buffer;
    }
}

void Renderer::print_program_info_log( GLuint obj) const
{
    int length;
    glGetProgramiv( obj, GL_INFO_LOG_LENGTH, &length);
    if( length > 1) {
        char* buffer = new char[length];
        int actual_length;
        glGetProgramInfoLog( obj, length, &actual_length, buffer);
        if( actual_length > 1)
            std::cerr << buffer << std::endl;
        delete[] buffer;
    }
}

#endif // HAVE_GLEW
